{     ,              ,        ( .., https://vk.com/wall162891802_892) }
{      ,      BOINC,       }
{ (c)   , 14.07.2019 }
program CellsMappingSchemas;
{$APPTYPE CONSOLE}

uses
  Windows;

const
  N = 8;


type
  TLatinSquare = array of array of Integer;
  TCellsMappingSchema = array [0..N*N-1] of Integer; {        }
  TUsedPairs = array [0..N*N-1] of Boolean;
  TNIntArr = array [0..N-1] of Boolean;
  TAllowedElemsArr = array [0..N-1] of TNIntArr;


procedure IndexToCoords(Index: Integer; var I, J: Integer);
begin
  I := Index div N;
  J := Index mod N;
end;


function CoordsToIndex(I, J: Integer): Integer;
begin
  Result := I*N + J;
end;


function GetCellsMappingSchemaForSodls(): TCellsMappingSchema;
var
  I, J, Index1, Index2: Integer;
begin
  for I := 0 to N-1 do
    for J := I to N-1 do begin
      Index1 := CoordsToIndex(I, J);
      Index2 := CoordsToIndex(J, I);

      Result[Index1] := Index2;
      Result[Index2] := Index1;
    end;
end;


{      : N      ( ),     = 2 }
function GetRandomCellsMappingSchemaForOdls(): TCellsMappingSchema;
var
  I, Index, I1, I2: Integer;
  NotUsed: TCellsMappingSchema;

  procedure DeleteValueFromNotUsed(Value: Integer);
  var
    Pos, Pos2, I: Integer;
  begin
    {   }
    Pos := -1;

    for I := 0 to High(NotUsed) do
      if NotUsed[I] = Value then begin
        Pos := I;
        break;
      end;

    {     >= 0 }
    Pos2 := -1;

    for I := High(NotUsed) downto 0 do
      if NotUsed[I] >= 0 then begin
        Pos2 := I;
        break;
      end;

    ASSERT(Pos >= 0);
    ASSERT(Pos2 >= 0);

    {   }
    NotUsed[Pos] := NotUsed[Pos2];
    NotUsed[Pos2] := -1;
  end;

begin
  for I := 0 to High(Result) do
    Result[I] := -1;

  {        }
  for I := 0 to N-1 do
    Result[I] := I;

  for I := 0 to N*N-1-N do
    NotUsed[I] := I + N;

  for I := N*N-N to N*N-1 do
    NotUsed[I] := -1;

  {         }
  for I := 0 to (N*N - N) div 2 - 1 do begin
    {   }
    Index := Random(N*N - N - I*2 - 1) + 1;

    ASSERT(NotUsed[Index] >= 0);

    {    2 }
    I1 := NotUsed[Index];
    I2 := NotUsed[0];

    Result[I1] := I2;
    Result[I2] := I1;

    {     NotUsed }
    DeleteValueFromNotUsed(I1);
    DeleteValueFromNotUsed(I2);
  end;
end;


function IsAllLsFilled(const LS: TLatinSquare): Boolean;
var
  I, J: Integer;
begin
  Result := False;

  for I := 0 to N-1 do
    for J := 0 to N-1 do
      if LS[I][J] < 0 then
        exit;

  Result := True;
end;


procedure PrintSquare(const LS: TLatinSquare);
var
  I, J: Integer;
begin
  for I := 0 to High(LS) do begin
    for J := 0 to High(LS) do
      if LS[I][J] >= 0 then
        Write(LS[I][J]:2)
      else
        Write(' -');
    Writeln;
  end;
end;


procedure SetLsItem_Fast(var LS: TLatinSquare; I, J, Value: Integer; var Rows, Strs: TAllowedElemsArr; var MainDiagonal, SecondDiagonal: TNIntArr);
begin
  LS[I][J] := Value;

  Rows[J][Value] := False;  {  }
  Strs[I][Value] := False;  {  }

  {  }
  if I = J then
    MainDiagonal[Value] := False;

  if I+J = N-1 then
    SecondDiagonal[Value] := False;
end;


procedure UnsetLsItem_Fast(var LS: TLatinSquare; I, J: Integer; var Rows, Strs: TAllowedElemsArr; var MainDiagonal, SecondDiagonal: TNIntArr);
var
  Value: Integer;
begin
  Value := LS[I][J];

  if Value >= 0 then begin
    Rows[J][Value] := True;  {  }
    Strs[I][Value] := True;  {  }

    {  }
    if I = J then
      MainDiagonal[Value] := True;

    if I+J = N-1 then
      SecondDiagonal[Value] := True;
  end;    

  LS[I][J] := -1;
end;


procedure Reset(var LS: TLatinSquare; FillFirstStringNormalized: Boolean = True);
var
  I, J: Integer;
begin
  {      }
  for I := 0 to High(LS) do
    for J := 0 to High(LS) do
      LS[I][J] := -1;

  {         }
  if FillFirstStringNormalized then
    for I := 0 to High(LS) do
      LS[0][I] := I;
end;


procedure Allocate(N: Integer; var LS: TLatinSquare);
begin
  SetLength(LS, N, N);
  Reset(LS);
end;


procedure Reset_Fast_CommonType(var Rows, Strs: TAllowedElemsArr; var MainDiagonal, SecondDiagonal: TNIntArr);
var
  I, J: Integer;
begin
  {      }
  for I := 0 to N-1 do
    MainDiagonal[I] := True;

  {      }
  for I := 0 to N-1 do
    SecondDiagonal[I] := True;

  {  }
  for I := 0 to N-1 do
    for J := 0 to N-1 do
      Strs[I][J] := True;

  {  }
  for I := 0 to N-1 do
    for J := 0 to N-1 do
      Rows[I][J] := True;
end;


procedure FillRowsAndStrsForPartialFilledSquare(const LS: TLatinSquare; var Rows, Strs: TAllowedElemsArr; var MD, SD: TNIntArr);
var
  I, J, Item: Integer;
begin
  Reset_Fast_CommonType(Rows, Strs, MD, SD);

  {   }
  for I := 0 to N-1 do begin
    Item := LS[I][I];

    if Item < 0 then
      continue;

    MD[Item] := False;
  end;

  {   }
  for I := 0 to N-1 do begin
    Item := LS[I][N-I-1];

    if Item < 0 then
      continue;

    SD[Item] := False;
  end;

  {  }
  for I := 0 to N-1 do
    for J := 0 to N-1 do begin
      Item := LS[I][J];

      if Item < 0 then
        continue;

      Strs[I][Item] := False;
    end;

  {  }
  for J := 0 to N-1 do
    for I := 0 to N-1 do begin
      Item := LS[I][J];

      if Item < 0 then
        continue;

      Rows[J][Item] := False;
    end;
end;


var
  BegValue, EndValue: Int64;
  ProcFreq: Int64 = -1;
  

function GetTickCountEx(): Int64; register;
asm
        rdtsc

        {$IFDEF Win64}
        shl rdx, 32
        add rax, rdx
        {$ENDIF}
end;


procedure StartProf();
begin
  BegValue := GetTickCountEx();
end;


procedure StopProf();
begin
  EndValue := GetTickCountEx();
end;


function ProfResult(): Int64;
begin
  Result := EndValue - BegValue;
end;


procedure MeasureCpuFreq();
const
  GET_FREQ_TIME_INTERVAL = 1000;
begin
  StartProf();
  Sleep(GET_FREQ_TIME_INTERVAL);
  StopProf();

  ProcFreq := Round(ProfResult() * 1000.0 / GET_FREQ_TIME_INTERVAL);
end;


procedure BruteForce_CellsMappingSchema(const CMS: TCellsMappingSchema);
var
  LS1, LS2: TLatinSquare;
  BF_Rows_LS1, BF_Strs_LS1: TAllowedElemsArr;
  BF_MainDiagonal_LS1, BF_SecondDiagonal_LS1: TNIntArr;
  BF_Rows_LS2, BF_Strs_LS2: TAllowedElemsArr;
  BF_MainDiagonal_LS2, BF_SecondDiagonal_LS2: TNIntArr;
  UsedPairs: array [0..N*N-1] of Boolean;
  I, Cnt: Integer;

  function GetUsedPairs(const LS1, LS2: TLatinSquare): TUsedPairs;
  var
    I, J: Integer;
  begin
    for I := 0 to N*N-1 do
      Result[I] := False;

    for I := 0 to N-1 do
      for J := 0 to N-1 do
        if (LS1[I][J] >= 0) and (LS2[I][J] >= 0) then
          Result[CoordsToIndex(LS1[I][J], LS2[I][J])] := True;
  end;

  function GetAllowedItemsForOdlsPairCnt(const Strs1, Rows1: TAllowedElemsArr; const MainDiagonal1, SecondDiagonal1: TNIntArr; const Strs2, Rows2: TAllowedElemsArr; const MainDiagonal2, SecondDiagonal2: TNIntArr; const UsedPairs: TUsedPairs; I, J: Integer): Integer;
  var
    V1, V2: Integer;
  begin
    Result := 0;

    for V1 := 0 to N-1 do
      if
        Strs1[I][V1] and                            {    }
        Rows1[J][V1] and                            {    }

        ((I <> J) or MainDiagonal1[V1]) and         {     }
        ((I+J <> N-1) or SecondDiagonal1[V1])       {     }
      then
        for V2 := 0 to N-1 do
          if
            Strs2[I][V2] and                            {    }
            Rows2[J][V2] and                            {    }

            ((I <> J) or MainDiagonal2[V2]) and         {     }
            ((I+J <> N-1) or SecondDiagonal2[V2]) and   {     }

            not UsedPairs[CoordsToIndex(V1, V2)]        {       }
          then
            Inc(Result);
  end;

  {           (v1,v2)   LS1  LS2 }
  procedure GetNextIndexes_MinCardFill_CMS(LS1, LS2: TLatinSquare; var NextX, NextY: Integer);
  var
    I, J, MinCard, ItemsCnt: Integer;
    UsedPairs: TUsedPairs;
  begin
    NextX := -1;
    NextY := -1;
    MinCard := N*N+1;

    UsedPairs := GetUsedPairs(LS1, LS2);

    for I := 0 to N-1 do
      for J := 0 to N-1 do begin
        if LS1[I][J] >= 0 then
          continue;

        ItemsCnt := GetAllowedItemsForOdlsPairCnt(BF_Strs_LS1, BF_Rows_LS1, BF_MainDiagonal_LS1, BF_SecondDiagonal_LS1, BF_Strs_LS2, BF_Rows_LS2, BF_MainDiagonal_LS2, BF_SecondDiagonal_LS2, UsedPairs, I, J);

        if ItemsCnt < MinCard then begin
          MinCard := ItemsCnt;
          NextX := I;
          NextY := J;

          if ItemsCnt <= 1 then
            exit;
        end;
      end;
  end;

  procedure Recurse(X, Y: Integer);
  var
    K1, K2, Index1, Index2: Integer;
    OrtX, OrtY: Integer;  {  ,     [X,Y]->[OrtX,OrtY]  CMS }
    NextX, NextY: Integer;
  begin
    {    }
    if (X < 0) and (Y < 0) then begin
      {         }
      if not IsAllLsFilled(LS1) then
        exit;

      {   }
      PrintSquare(LS1);

      Inc(Cnt);
      StopProf();

      Writeln('ODLS found: ', Cnt,  ', pace: ', Cnt / (ProfResult()/ProcFreq):2:2, ' ODLS/s'#13#10);

      exit;
    end;

    {   [OrtX,OrtY],     [X,Y]  CMS }
    Index1 := CoordsToIndex(X, Y);
    Index2 := CMS[Index1];
    IndexToCoords(Index2, OrtX, OrtY);

    {     (K1,K2) }
    for K1 := 0 to N-1 do begin
      { LS1[X][Y] := K1 }
      {    LS1? }
      if not BF_Strs_LS1[X][K1] then
        continue;

      {    LS1? }
      if not BF_Rows_LS1[Y][K1] then
        continue;

      {     LS1? }
      if (X = Y) and not BF_MainDiagonal_LS1[K1] then
        continue;

      {     LS1? }
      if (X = N-1 - Y) and not BF_SecondDiagonal_LS1[K1] then
        continue;

      { LS2[OrtX][OrtY] := K1 }
      {    LS2? }
      if not BF_Strs_LS2[OrtX][K1] then
        continue;

      {    LS2? }
      if not BF_Rows_LS2[OrtY][K1] then
        continue;

      {     LS2? }
      if (OrtX = OrtY) and not BF_MainDiagonal_LS2[K1] then
        continue;

      {     LS2? }
      if (OrtX = N-1 - OrtY) and not BF_SecondDiagonal_LS2[K1] then
        continue;

      {     ,       // }  
      SetLsItem_Fast(LS1,    X,    Y, K1, BF_Rows_LS1, BF_Strs_LS1, BF_MainDiagonal_LS1, BF_SecondDiagonal_LS1);
      SetLsItem_Fast(LS2, OrtX, OrtY, K1, BF_Rows_LS2, BF_Strs_LS2, BF_MainDiagonal_LS2, BF_SecondDiagonal_LS2);

      for K2 := 0 to N-1 do begin
        { K1 <> K2, ..          }
        if K1 = K2 then
          continue;

        {   (K1,K2)    }
        if UsedPairs[CoordsToIndex(K1,K2)] then
          continue;

        {  (K2,K1)     }
        ASSERT( not UsedPairs[CoordsToIndex(K2,K1)] );

        { LS1[OrtX][OrtY] := K2 }
        {    LS1? }
        if not BF_Strs_LS1[OrtX][K2] then
          continue;

        {    LS1? }
        if not BF_Rows_LS1[OrtY][K2] then
          continue;

        {     LS1? }
        if (OrtX = OrtY) and not BF_MainDiagonal_LS1[K2] then
          continue;

        {     LS1? }
        if (OrtX = N-1 - OrtY) and not BF_SecondDiagonal_LS1[K2] then
          continue;

        { LS2[X][Y] := K2 }
        {    LS2? }
        if not BF_Strs_LS2[X][K2] then
          continue;

        {    LS2? }
        if not BF_Rows_LS2[Y][K2] then
          continue;

        {     LS2? }
        if (X = Y) and not BF_MainDiagonal_LS2[K2] then
          continue;

        {     LS2? }
        if (X = N-1 - Y) and not BF_SecondDiagonal_LS2[K2] then
          continue;

        {   :   }
        SetLsItem_Fast(LS1, OrtX, OrtY, K2, BF_Rows_LS1, BF_Strs_LS1, BF_MainDiagonal_LS1, BF_SecondDiagonal_LS1);
        SetLsItem_Fast(LS2,    X,    Y, K2, BF_Rows_LS2, BF_Strs_LS2, BF_MainDiagonal_LS2, BF_SecondDiagonal_LS2);

        UsedPairs[CoordsToIndex(K1, K2)] := True;
        UsedPairs[CoordsToIndex(K2, K1)] := True;

        {        http://evatutin.narod.ru/evatutin_co_ls_dls_1_8.pdf }
        GetNextIndexes_MinCardFill_CMS(LS1, LS2, NextX, NextY);

        Recurse(NextX, NextY);

        UnsetLsItem_Fast(LS1, OrtX, OrtY, BF_Rows_LS1, BF_Strs_LS1, BF_MainDiagonal_LS1, BF_SecondDiagonal_LS1);
        UnsetLsItem_Fast(LS2,    X,    Y, BF_Rows_LS2, BF_Strs_LS2, BF_MainDiagonal_LS2, BF_SecondDiagonal_LS2);

        UsedPairs[CoordsToIndex(K1, K2)] := False;
        UsedPairs[CoordsToIndex(K2, K1)] := False;
      end;

      UnsetLsItem_Fast(LS1,    X,    Y, BF_Rows_LS1, BF_Strs_LS1, BF_MainDiagonal_LS1, BF_SecondDiagonal_LS1);
      UnsetLsItem_Fast(LS2, OrtX, OrtY, BF_Rows_LS2, BF_Strs_LS2, BF_MainDiagonal_LS2, BF_SecondDiagonal_LS2);
    end;
  end;

var
  X, Y, CurrVal: Integer;

begin
  StartProf();

  Allocate(N, LS1);
  Allocate(N, LS2);

  {    ,          (, SODLS -   ) }
  Reset(LS1, False);
  Reset(LS2, False);

  CurrVal := 0;

  for I := 0 to N*N-1 do
    if CMS[I] = I then begin
      IndexToCoords(I, X, Y);

      LS1[X][Y] := CurrVal;
      LS2[X][Y] := CurrVal;

      Inc(CurrVal);
    end;

  { CurrVal      N,      ,   N, ..    }
  ASSERT(CurrVal = N);

  FillRowsAndStrsForPartialFilledSquare(LS1, BF_Rows_LS1, BF_Strs_LS1, BF_MainDiagonal_LS1, BF_SecondDiagonal_LS1);
  FillRowsAndStrsForPartialFilledSquare(LS2, BF_Rows_LS2, BF_Strs_LS2, BF_MainDiagonal_LS2, BF_SecondDiagonal_LS2);

  {     }
  for I := 0 to N*N-1 do
    UsedPairs[I] := False;

  {   (I,I) }
  for I := 0 to N-1 do
    UsedPairs[CoordsToIndex(I,I)] := True;

  Cnt := 0;

  Recurse(1, 0);
end;


var
  CMS: TCellsMappingSchema;
  Choice: Integer;

begin
  Writeln('Measuring CPU frequency...');

  MeasureCpuFreq();

  Writeln('Current CPU frequency is ', ProcFreq/1e9:2:2, ' GHz');

  Writeln(#13#10'Select the schema: ');
  Writeln(' 1 - random');
  Writeln(' 2 - ODLS');
  Writeln(' 3 - classical SODLS');

  Write(#13#10'Schema: ');
  Readln(Choice);

  case Choice of
    1: begin
      {  1   }
      //RandSeed := 0;

      Randomize();
      CMS := GetRandomCellsMappingSchemaForOdls();
    end;

    2: begin
      {  2   }
      CMS[0] := 0;
      CMS[1] := 1;
      CMS[2] := 2;
      CMS[3] := 3;
      CMS[4] := 4;
      CMS[5] := 5;
      CMS[6] := 6;
      CMS[7] := 7;
      CMS[8] := 45;
      CMS[9] := 40;
      CMS[10] := 41;
      CMS[11] := 44;
      CMS[12] := 36;
      CMS[13] := 47;
      CMS[14] := 20;
      CMS[15] := 54;
      CMS[16] := 46;
      CMS[17] := 59;
      CMS[18] := 55;
      CMS[19] := 34;
      CMS[20] := 14;
      CMS[21] := 38;
      CMS[22] := 23;
      CMS[23] := 22;
      CMS[24] := 31;
      CMS[25] := 27;
      CMS[26] := 29;
      CMS[27] := 25;
      CMS[28] := 30;
      CMS[29] := 26;
      CMS[30] := 28;
      CMS[31] := 24;
      CMS[32] := 56;
      CMS[33] := 42;
      CMS[34] := 19;
      CMS[35] := 51;
      CMS[36] := 12;
      CMS[37] := 52;
      CMS[38] := 21;
      CMS[39] := 63;
      CMS[40] := 9;
      CMS[41] := 10;
      CMS[42] := 33;
      CMS[43] := 49;
      CMS[44] := 11;
      CMS[45] := 8;
      CMS[46] := 16;
      CMS[47] := 13;
      CMS[48] := 53;
      CMS[49] := 43;
      CMS[50] := 60;
      CMS[51] := 35;
      CMS[52] := 37;
      CMS[53] := 48;
      CMS[54] := 15;
      CMS[55] := 18;
      CMS[56] := 32;
      CMS[57] := 58;
      CMS[58] := 57;
      CMS[59] := 17;
      CMS[60] := 50;
      CMS[61] := 62;
      CMS[62] := 61;
      CMS[63] := 39;
    end;

    3:
      {  3  SODLS }
      CMS := GetCellsMappingSchemaForSodls();

    else begin
      Writeln('Selected schema is unsupported');
      Readln;
      Halt;
    end;
  end;  

  BruteForce_CellsMappingSchema(CMS);

  Writeln('OK');
  Readln;
end.
